/* 

Copyright (C) 2003-2004 Yves Schmid (www.garagecube.com)

Author(s): Yves Schmid

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     GNU General Public License for more details.

You should have received a copy of the GNU General Public Licensealong with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#import "CosmicOpenGLView.h"
#import "CosmicPalette.h"
#import "CosmicDocument.h"
#import "CosmicMirror.h"

#include "OMediaWorld.h"
#include "OMediaViewPort.h"
#include "OMediaLayer.h"
#include "OMediaGLRenderer.h"
#include "OMedia3DMaterial.h"
#include "OMediaEngineFactory.h"
#include "OMediaCanvas.h"
#include "OMediaCanvasElement.h"
#include "OMediaTimer.h"
#include "OMediaPickRequest.h"
#include "OMediaMatrix.h"
#include "OMediaString.h"

#import <OpenGL/gl.h>
#import <OpenGL/glext.h>
#import <OpenGL/OpenGL.h>

#include <vector>

#define		kVideoTimeScale 	1000

#define COSMIC_MIN_Z -2200.0f
#define COSMIC_MAX_Z 288.0f

extern BOOL cosmicStopAllThread;
extern int  cosmitThreadsRunning;

static BOOL osNeedsRedraw;

static NSLock *renderLock;

class CosmicGLRenderTarget : public OMediaPureGLRenderTarget
{
    public:

    CosmicOpenGLView	*glView;

    CosmicGLRenderTarget() {glView=NULL;}

	virtual void enter_context(void)
	{
		if (!renderLock)
			renderLock = [[NSRecursiveLock alloc] init];

		[renderLock lock];

		if (glView) [[glView openGLContext] makeCurrentContext];
	}

	virtual void exit_context(void)
	{
		if (renderLock)
			[renderLock unlock];
	}
};

@implementation CosmicOpenGLView

- (BOOL)isOpaque
{
    return YES;
}

- (void)lock
{
	if (!renderLock)
		renderLock = [[NSRecursiveLock alloc] init];
	[renderLock lock];
}

- (void)unlock
{
	[renderLock unlock];
}

-(NSString*)floatToStringRound:(float)val
{
    NSDecimal	src,dest;
    NSScanner	*scanner;

    scanner = [NSScanner scannerWithString:[NSString stringWithFormat:@"%f",val]];
    
    if (![scanner scanDecimal:&src]) return @"0";

    NSDecimalRound(&dest, &src, 2, NSRoundPlain);

    return NSDecimalString(&dest, NULL);    
}

- (void)awakeFromNib
{
	[[self window] setAspectRatio:NSMakeSize(1,1)];
}

- (void)cosmicInit
{
	if (initialized) return;

    long params[] = { 1};

    [[self openGLContext] makeCurrentContext];

    renderTarget = new CosmicGLRenderTarget;
    ((CosmicGLRenderTarget*)renderTarget)->glView = self;

	CGLSetParameter(CGLGetCurrentContext(),  kCGLCPSwapInterval, params);

	palette = [CosmicPalette getMain];
	

	OMediaFARGBColor	argb;

	world = new OMediaWorld;
	world->set_input_engine(OMediaEngineFactory::get_factory()->create_input_engine(ommeic_OS,NULL));
	
	viewport = new OMediaViewPort;
	viewport->link(world); 
	viewport->place(0,0,0);
	
	argb.set(1.0f,1.0f,0.0f,0);

	layer = new OMediaLayer;
	layer->link(world);
	layer->set_flags(/*omlayerf_ClearColor|*/omlayerf_Visible);
	layer->set_clear_color(argb);
	layer->set_projection_type(omlptc_PerspectiveFOV);

	argb.set(1.0f,0.0f,0.0f,0.0f);
	
	canvas = new OMediaCanvas;
	canvas->create(512,512);
	canvas->fill(argb);
	canvas->set_filtering(omtfc_Linear,omtfc_Linear);

	undoBuffer = new unsigned long[canvas->get_width()*canvas->get_height()];
	
	for(int gy=0;gy<COSMIC_GRID_SIZE;gy++)
	for(int gx=0;gx<COSMIC_GRID_SIZE;gx++)
	{
		canvasElement[gx][gy] = new OMediaCanvasElement;
		canvasElement[gx][gy]->link(world);
		canvasElement[gx][gy]->set_auto_align(omaac_Center,omaac_Center);
		canvasElement[gx][gy]->set_canvas(canvas);
		canvasElement[gx][gy]->place(gx*canvas->get_width(),gy*canvas->get_height(),310);
	}
	
	argb.set(0.0f,0.0f,0.0f,0.0f);
	crossCanvas = new OMediaCanvas;
	crossCanvas->create(64,64);
	crossCanvas->fill(argb);
	argb.set(0.5f,1.0f,1.0f,1.0f);
	crossCanvas->draw_line(argb,31,0,31,31);
	crossCanvas->draw_line(argb,0,31,31,31);
	argb.set(0.5f,0.0f,0.0f,0.0f);
	crossCanvas->draw_line(argb,32,0,32,32);
	crossCanvas->draw_line(argb,0,32,32,32);

	crossCanvasElement = new OMediaCanvasElement;
	crossCanvasElement->link(world);
	crossCanvasElement->set_auto_align(omaac_Center,omaac_Center);
	crossCanvasElement->set_canvas(crossCanvas);
	crossCanvasElement->place(0,0,308);
	crossCanvasElement->set_blend(omblendfc_Src_Alpha,omblendfc_Inv_Src_Alpha);
	crossCanvasElement->set_flags(omelf_DisablePicking|omelf_DisableViewportTransform);
	crossCanvasElement->hide();

	argb.set(0.0f,0.0f,0.0f,0.0f);
	bufferBorderCanvas = new OMediaCanvas;
	bufferBorderCanvas->create(2,2);
	bufferBorderCanvas->lock(omlf_Write);
	bufferBorderCanvas->write_pixel(OMediaFARGBColor(1.0,1.0,1.0,1.0),0,0);
	bufferBorderCanvas->write_pixel(OMediaFARGBColor(1.0,0.6,0.6,0.6),1,0);
	bufferBorderCanvas->write_pixel(OMediaFARGBColor(1.0,0.6,0.6,0.6),0,1);
	bufferBorderCanvas->write_pixel(OMediaFARGBColor(1.0,0.3,0.3,0.3),1,1);
	bufferBorderCanvas->unlock();
	bufferBorderCanvas->set_filtering(omtfc_Linear,omtfc_Linear);

	bufferCanvasElement = new OMediaCanvasElement;
	bufferCanvasElement->link(world);
	bufferCanvasElement->set_auto_align(omaac_Center,omaac_Center);
	bufferCanvasElement->set_canvas(canvas);
	bufferCanvasElement->place(128,-128,308);
	bufferCanvasElement->set_flags(omelf_DisablePicking|omelf_DisableViewportTransform);
	bufferCanvasElement->set_canvas_flags(omcanef_FreeWorldSize);
	bufferCanvasElement->set_size(64,64);
	bufferCanvasElement->hide();
	
	OMediaCanvasElement *bufferBorder = new OMediaCanvasElement;
	bufferBorder->link(bufferCanvasElement);
	bufferBorder->set_auto_align(omaac_Center,omaac_Center);
	bufferBorder->set_canvas(bufferBorderCanvas);
	bufferBorder->place(0,0,0);
	bufferBorder->set_size(66,66);
	bufferBorder->set_canvas_flags(omcanef_FreeWorldSize);
	bufferBorder->set_flags(omelf_DisablePicking|omelf_DisableViewportTransform);

	rotationSpeed = [palette->rotationSpeedSlider floatValue];
	rotationSpeed = omd_Deg2AngleF(rotationSpeed);
	
	scrollSpeed = [palette->scrollSpeedSlider floatValue];
	scaleSpeed = [palette->scaleSpeedSlider floatValue];
	scaleStart = [palette->scaleStartTextField floatValue]/100.0f;
	scaleEnd  = [palette->scaleEndTextField floatValue]/100.0f;
	if (scaleStart<0) scaleStart = 0;
	else if (scaleStart>1.0f) scaleStart = 1.0f;
	if (scaleEnd<0) scaleEnd = 0;
	else if (scaleEnd>1.0f) scaleEnd = 1.0f;	
	if (scaleEnd<scaleStart) scaleEnd = scaleStart;
	scalePingpongReverse = 1.0f;

	scalePingpong = [palette->scalePingpongCheckbox intValue]!=0;

	wantedRate = 1.0/40.0;
	fpsFactor = [palette->fpsSlider floatValue];

	connectLines = FALSE;
	mouseDownFirstDraw = YES;

	[NSThread detachNewThreadSelector:@selector(renderThread:) toTarget:self withObject:nil]; 

    mainThreadTimer = [[NSTimer scheduledTimerWithTimeInterval:1.0/50.0 target:self selector:@selector(mainThreadTimer:) userInfo:nil repeats:YES] retain];
    updateUITimer = [[NSTimer scheduledTimerWithTimeInterval:1.0/5.0 target:self selector:@selector(updateUITimer:) userInfo:nil repeats:YES] retain];

	initialized = TRUE;
}

- (id)initWithFrame:(NSRect)frameRect //pixelFormat:(NSOpenGLPixelFormat *)format
{	
	NSOpenGLPixelFormatAttribute attrs[] =
	{
		NSOpenGLPFANoRecovery,
		NSOpenGLPFAAccelerated,
		NSOpenGLPFADoubleBuffer,
	//	NSOpenGLPFADepthSize, 16,
		(NSOpenGLPixelFormatAttribute)0
	};
	
	NSOpenGLPixelFormat* pixFmt = [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];
	if(!pixFmt)
	{
		NSLog(@"No pixel format -- exiting");
		exit(1);
	}
		
	self = [super initWithFrame:frameRect pixelFormat: pixFmt];

	[self cosmicInit];
	return self;
}

-(void)flushCanvasGLTexture
{
	if (!canvas) return;
	canvas->delete_imp_slaves();
	crossCanvas->delete_imp_slaves();
	bufferBorderCanvas->delete_imp_slaves();
}

-(void)stopAndReleaseAll
{
	if (renderTarget==NULL) return;

	stopAll = YES;
	while(!threadTerminated) [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0/20.0]];
	
	[mainThreadTimer invalidate];
	[updateUITimer invalidate];

	if ([CosmicMirror getViewToMirror]==self)
		[CosmicMirror setViewToMirror:nil];

    [[self openGLContext] makeCurrentContext];	// Always make current context before freeing...
    
    delete world;
    delete renderTarget;
	delete canvas;
	delete crossCanvas;
	delete bufferBorderCanvas;
	delete [] undoBuffer;

	canvas = NULL;
	renderTarget = NULL;
}

-(void)dealloc
{
	[self finishMovieCompression];

	[self stopAndReleaseAll];
	[super dealloc];
}

- (void)processColorInterpolation:(float)elapsed
{
	float   speed = [palette->colorSpeedSlider floatValue]*5.0f;

	colorInterpolateState += (elapsed/1000.0f) * speed;
	
	if (colorInterpolateState>1.0f) 
	{
		if ([palette->colorCircularSlider state]!=NSOnState)
			colorInterpolateState = 1.0f;
		else
			while(colorInterpolateState>1.0f) colorInterpolateState -= 1.0f;
	}
	
	if (mouseIsDown)
		[palette rebuildWithColorState:colorInterpolateState];
}

- (float)mouseToRotate
{
	OMedia3DVector  v;
	NSPoint p = [[self window] mouseLocationOutsideOfEventStream];
	float   angle;
		
	angle = omd_ATan2(([self bounds].size.width*0.5f)-p.x,([self bounds].size.height*0.5f)-p.y);

	return angle;
}


-(float)findViewWorldFactor
{
	OMediaMatrix_4x4  proj_matrix;
	OMedia3DPoint   p;

	float   ratio = [self bounds].size.height/[self bounds].size.width;
	
	proj_matrix.set_perspective_fov_zn1_1(layer->get_proj_perspective_fov(),
											layer->get_near_clip(),layer->get_far_clip(),
											ratio);

	p.set(100.0,0,canvasElement[0][0]->getz()-viewport->getz());
	proj_matrix.multiply(p);
	
	p.x = (p.x / p.w);
	p.x *= [self bounds].size.width*0.5;
	
	return 100.0/p.x;
}

-(void)realignViewPort
{
	float			vx,vy,w,h;
	int				iw,ih;
	bool			gmoved = false;

	vx = viewport->getx();
	vy = viewport->gety();

	w = canvas->get_width();
	h = canvas->get_height();
	
	iw = (int)w;
	ih = (int)h;
		
	while (vx<canvasElement[(COSMIC_GRID_SIZE/2)-1][(COSMIC_GRID_SIZE/2)-1]->getx()) {vx += w; lastEX+=iw; gmoved=true;}   // Ugly "while" - could be optimized
	while (vx>canvasElement[(COSMIC_GRID_SIZE/2)][(COSMIC_GRID_SIZE/2)]->getx()) {vx -= w; lastEX-=iw; gmoved=true;} 

	while (vy<canvasElement[(COSMIC_GRID_SIZE/2)-1][(COSMIC_GRID_SIZE/2)-1]->gety()) {vy += h;lastEY+=ih;gmoved=true;}
	while (vy>canvasElement[(COSMIC_GRID_SIZE/2)][(COSMIC_GRID_SIZE/2)]->gety()) {vy -= h; lastEY-=ih;gmoved=true;}

	if (gmoved)  lastPickedElement=NULL;
		
	viewport->place(vx,vy);	
}

-(void)moveViewX:(float)x Y:(float)y viewCoords:(BOOL)viewCoords
{
	OMedia3DVector  v;
	float			f = (viewCoords)?[self findViewWorldFactor]:1.0f;
	
	v.x = -x * f;
	v.y = -y * f;
	v.z = 0;
	
	v.rotate(0,0,viewport->get_anglez());

	viewport->move(v);
	
	// Realign on grid
	
	[self realignViewPort];
		
}

-(void)applyScaling:(float)elapsed
{
	float   z,maxz,minz;

	if (scaleSpeed)
	{	
		z = scaleSpeed * elapsed * 10;

		if (scalePingpong) z *= scalePingpongReverse;

		viewport->move(0,0,z);
		

		minz = COSMIC_MIN_Z + (COSMIC_MAX_Z-COSMIC_MIN_Z)*(1.0f-scaleEnd);
		maxz = COSMIC_MIN_Z + (COSMIC_MAX_Z-COSMIC_MIN_Z)*(1.0f-scaleStart);
		
		if (scalePingpong)
		{
			if (viewport->getz()>maxz) 
			{
				viewport->setz(maxz);
				scalePingpongReverse = -scalePingpongReverse;
			}
			else if (viewport->getz()<minz) 
			{
				viewport->setz(minz);
				scalePingpongReverse = -scalePingpongReverse;
			}
		}
		else
		{
			if (viewport->getz()>maxz) viewport->setz(minz);
			else if (viewport->getz()<minz) viewport->setz(maxz);
		}
	}
}

-(void)applyScrolling:(float)elapsed
{
	OMedia3DVector  v(0,-elapsed * scrollSpeed*10,0);
	
	v.rotate(0,0,viewport->get_anglez());
	[self moveViewX:v.x Y:v.y viewCoords:NO];
}

- (void)periodical
{
	if (!renderTarget) return;

	CosmicDocument  *document = (CosmicDocument*)[[[self window] windowController] document];
	if (document)
	{
		// Update document changes if required...
	
		if (document->setCanvas)
		{
			canvas->draw_full(document->setCanvas,0,0);
			delete document->setCanvas;
			document->setCanvas = NULL;

			rotationSpeed = document->setRotateSpeed;
			fpsFactor = document->setFPSFactor;
			
			viewport->place(document->setPX,document->setPY,document->setPZ);
			viewport->set_angle(document->setAX,document->setAY,document->setAZ);
			scrollSpeed = document->setScrollSpeed;
			scaleSpeed = document->setScaleSpeed;
			scaleStart = document->setScaleStart;
			scaleEnd = document->setScaleEnd;
			scalePingpong = document->setScalePingpong;

			[self realignViewPort];
			keyWindowChanged = true;
		}
	}


	BOOL	paintQuiet = !mouseIsDown;
	float   elapsed = (float)(wantedRate*1000.0)/fpsFactor;

	if ((palette->currentTool!=cosmic_Rotate && 
			palette->currentTool!=cosmic_Scroll && 
			palette->currentTool!=cosmic_Scale) || !mouseIsDown)
	{
		viewport->add_angle(0,0,elapsed * -rotationSpeed);		
		[self applyScrolling:elapsed];
		[self applyScaling:elapsed];
	}
	
	[self processColorInterpolation:elapsed];

	if (palette->currentTool==cosmic_Rotate)
	{	
		if (mouseIsDown)
			handRotate = [self mouseToRotate]-baseHandRotate;
		else
		{
			if (handRotate!=0)
			{
				viewport->add_angle(0,0,-handRotate);
				baseHandRotate = 0;
				handRotate = 0;
			}
		}	
		
		paintQuiet = YES;
	}
	else if (palette->currentTool==cosmic_Scroll)
	{
		if (mouseIsDown)
		{
			NSPoint mp = [NSEvent mouseLocation];
			
			[self moveViewX:mp.x - baseHandMoveX Y:mp.y - baseHandMoveY viewCoords:YES];
			baseHandMoveX = mp.x;
			baseHandMoveY = mp.y;
		}

		paintQuiet = YES;
	}
	else if (palette->currentTool==cosmic_Scale)
	{
		if (mouseIsDown)
		{		
			NSPoint mp = [NSEvent mouseLocation];
			
			tempHandScale = mp.y - baseHandScaleY;
			if (tempHandScale>0)
				tempHandScale *= 1.0f/50.0f;
			else
				tempHandScale *= 1.0f/100.0f;
		}

		paintQuiet = YES;
	}

	if (!mouseIsDown && tempHandScale)
	{	
		viewport->move(0,0,tempHandScale*200);
		if (viewport->getz()>COSMIC_MAX_Z) viewport->setz(COSMIC_MAX_Z);
		else if (viewport->getz()<COSMIC_MIN_Z) viewport->setz(COSMIC_MIN_Z);
		tempHandScale = 0;  
		baseHandScaleY = 0;
	}			
	
	[self paintBrush:paintQuiet];
}


- (void)render
{
    NSRect rect;
    OMediaRect	vp_bounds;
	float		saveZ;
	
	[CosmicMirror mirrorRenderView:self];

    [[self openGLContext] makeCurrentContext];

	if ([palette->crossCheckbox state]==NSOnState)
		crossCanvasElement->force_show();
	else
		crossCanvasElement->hide();

	if ([palette->bufferCheckbox state]==NSOnState)
		bufferCanvasElement->force_show();
	else
		bufferCanvasElement->hide();

    rect = [self bounds];

    vp_bounds.set(0,0, (long) rect.size.width, (long) rect.size.height);
    viewport->setbounds(&vp_bounds);

	float tempRotate = mouseIsDown?handRotate:0;

	if (palette->currentTool==cosmic_Rotate)
		viewport->add_angle(0,0,-tempRotate);

	saveZ = viewport->getz();

	if (palette->currentTool==cosmic_Scale)
	{
		viewport->move(0,0,tempHandScale*200);
		if (viewport->getz()>COSMIC_MAX_Z) viewport->setz(COSMIC_MAX_Z);
		else if (viewport->getz()<COSMIC_MIN_Z) viewport->setz(COSMIC_MIN_Z);
	}

    world->update_logic();
    renderTarget->render_viewport(viewport);

	if (palette->currentTool==cosmic_Rotate)
		viewport->add_angle(0,0,tempRotate);

	if (palette->currentTool==cosmic_Scale)
		viewport->setz(saveZ);
	
    // Swap buffer to screen
    [[self openGLContext] flushBuffer];
}

- (void)renderThread:(id)anObject
{
    NSAutoreleasePool 		*pool = [[NSAutoreleasePool alloc] init];
    OMediaTimer				autoreleasePoolTimer,timer;

	//[NSThread setThreadPriority:0.7];

	cosmitThreadsRunning++;   

	while (!initialized)
	{
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0/20.0]];
	}
 
    autoreleasePoolTimer.start();
    
    while(!stopAll && !cosmicStopAllThread)
    {
	
		double idealTime = wantedRate * (1.0/fpsFactor);
	
		double toWait;
		double elapsed = timer.getelapsed();
								
		toWait = (idealTime)-(elapsed/1000.0);

		if (toWait>idealTime || toWait>1) toWait = 1;
		if (toWait<=1.0/100.0) toWait = 1.0/100.0;
			
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:toWait]];		
		timer.start();

        if (autoreleasePoolTimer.getelapsed()>1000)
        {
            [pool release];
            pool = [[NSAutoreleasePool alloc] init];
            autoreleasePoolTimer.start();
        }
		
		// Simple hack to stop the animation for un-selected windows
		static NSWindow *lastMainWindow;
		NSWindow		*mainWin = [[NSApplication sharedApplication] mainWindow]; 
		
		if (!mainWin) mainWin = lastMainWindow;
		
		if ( (mainWin!=[self window]) || recordMode)   // Does not work if application is in the background
		{
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0/30.0]];
			continue;
		}

		if (mainWin)
			lastMainWindow = mainWin;

		if (osNeedsRedraw)
		{
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0/30.0]];		
		}
		else if (renderLock)
		{
			if ([self lockFocusIfCanDraw])
			{			
				[self lock];
				[self periodical];
				[self render];
				
				[self unlockFocus];
				[self unlock];
			}
		}
    }
    
    
    [pool release];
    threadTerminated = YES;
	cosmitThreadsRunning--;   
}


- (void)drawRect:(NSRect)aRect
{
	if (!initialized) [self cosmicInit];
	osNeedsRedraw = YES;
	[self lock];
	[self render];
	[self unlock];
	osNeedsRedraw = NO;
}

- (void)mouseDown:(NSEvent *)theEvent
{
	NSPoint mp = [NSEvent mouseLocation];

	mouseDownFirstDraw = YES;
	mouseIsDown = YES;
	colorInterpolateState = 0.0f;
	baseHandRotate = [self mouseToRotate];
	handRotate = 0;
	baseHandMoveX = mp.x;
	baseHandMoveY = baseHandScaleY = mp.y;

	[[self window] invalidateCursorRectsForView:self];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	mouseIsDown = NO;
	mouseDownFirstDraw = YES;

	[[self window] invalidateCursorRectsForView:self];
}

- (void)directPaintBrushAtX:(int)x Y:(int)y
{
	OMediaCanvas	*srcCanv = [palette getCurrentBrush];

	if ([palette->drawShadowCheckbox state]==NSOnState)
		canvas->draw_full(palette->currentBrushShadow,x+4,y+4,omblendfc_Src_Alpha,omblendfc_Inv_Src_Alpha);

	canvas->draw_full(srcCanv,x,y,omblendfc_Src_Alpha,omblendfc_Inv_Src_Alpha);
}

- (void)paintBrushAtX:(int)x Y:(int)y
{
	if (!paintDisableAutoTiledDraw)
	{
		OMediaCanvas	*srcCanv = [palette getCurrentBrush];
		int				x2,y2;
		
		x2 = x+srcCanv->get_width()+4;
		y2 = y+srcCanv->get_height()+4;

		if (x<0 && y2>canvas->get_height()) [self directPaintBrushAtX:x+canvas->get_width() Y:y-canvas->get_height()];
		if (x2>canvas->get_width() && y<0) [self directPaintBrushAtX:x-canvas->get_width() Y:y+canvas->get_height()];

		if (x<0 && y<0) [self directPaintBrushAtX:x+canvas->get_width() Y:y+canvas->get_height()];
		if (x2>canvas->get_width() && y2>canvas->get_height()) [self directPaintBrushAtX:x-canvas->get_width() Y:y-canvas->get_height()];

		if (x<0) [self directPaintBrushAtX:x+canvas->get_width() Y:y];
		if (y<0) [self directPaintBrushAtX:x Y:y+canvas->get_height()];
		if (x2>canvas->get_width()) [self directPaintBrushAtX:x-canvas->get_width() Y:y];
		if (y2>canvas->get_height()) [self directPaintBrushAtX:x Y:y-canvas->get_height()];
	}

	[self directPaintBrushAtX:x Y:y];

}


- (void)paintBrushAFromX:(int)x1 Y:(int)y1 toX:(int)x2 toY:(int)y2  skipFirstPoint:(BOOL)skipFirstPoint
{
	int denominator, frac_x,frac_y,i,run,rise,sign_x,sign_y;
	int destx,desty;

	run = x2-x1;
	rise = y2-y1;

	destx = x1;
	desty = y1;

	if (run > 0) sign_x = 1;
	else
	{
		sign_x = -1;
		run = -run;
	}
	
	if (rise>0) sign_y = 1;
	else
	{
		sign_y = -1;
		rise = -rise;
	}
	
	denominator = (rise>run)?rise:run;
	frac_y = frac_x = denominator>>1;
	
	for (i=denominator;i;--i)
	{
		if (i!=denominator || !skipFirstPoint) // Skip first point
			[self paintBrushAtX:destx Y:desty];

		if ( (frac_x+=run) > denominator)
		{
			frac_x -=denominator;
			destx += sign_x;
		}
		
		if ( (frac_y+=rise) > denominator)
		{
			frac_y -=denominator;
			desty += sign_y;
		}
	}

	[self paintBrushAtX:destx Y:desty];
}

- (void)paintBrushAFromX:(int)x1 Y:(int)y1 toX:(int)x2 toY:(int)y2
{
	[self paintBrushAFromX:x1 Y:y1 toX:x2 toY:y2 skipFirstPoint:YES];
}

- (void)paintAcrossTilesToX:(int)x1 toY:(int)y1 element:(OMediaElement*)element
{
	int px,py;
	int w = canvas->get_width(),h = canvas->get_height();
	int x2=lastX,y2=lastY;
	int gw = canvas->get_width()*COSMIC_GRID_SIZE,gh = canvas->get_height()*COSMIC_GRID_SIZE;
	OMediaRect  r1,r2;
	int		px1,px2,py1,py2;
	OMediaCanvas	*srcCanv = [palette getCurrentBrush];
	int		brw,brh;
	if (!srcCanv) return;
	
	brw = srcCanv->get_width()>>1;
	brh = srcCanv->get_height()>>1;

	paintDisableAutoTiledDraw = YES;
			
	x1+=(int)element->getx();
	y1=((int)element->gety())-y1;

	x2+=lastEX;
	y2=lastEY-y2;

	r2.set(0,0,w,h);
		
	for(py=0;py<gw;py+=w)
	for(px=0;px<gh;px+=h)
	{
		px1 = x2-px;
		py1 = py-y2;
		px2 = x1-px;
		py2 = py-y1;
		
		if (px1<px2)
		{
			r1.left = px1-brw; r1.right = px2+brw;
		}
		else
		{
			r1.left = px2-brw; r1.right = px1+brw;
		}

		if (py1<py2)
		{
			r1.top = py1-brh; r1.bottom = py2+brh;
		}
		else
		{
			r1.top = py2-brh; r1.bottom = py1+brh;
		}
	
		if (r1.touch(&r2))
		{
			[self paintBrushAFromX:px1 Y:py1 toX:px2 toY:py2];
		}

	}

	paintDisableAutoTiledDraw = NO;
}

- (void)paintBrush:(BOOL)quiet
{
	NSPoint p = [[self window] mouseLocationOutsideOfEventStream];
	float   mx=p.x,my=([self bounds].size.height-p.y)+14;

	OMediaPickRequest   pr;
	OMediaFARGBColor	argb;
	
	pr.pickx = mx;
	pr.picky = my;
	pr.pickw = 0.5f;
	pr.pickh = 0.5f;
	pr.level = omplc_Surface;
	pr.flags = ompickf_CloserHit_SurfaceHitOnly;

	renderTarget->pick_viewport(viewport,&pr);
	
	if (pr.closer_hit.type!=omptc_Element || pr.closer_hit.element==NULL) return;
	if (pr.closer_hit.sub_info.size()==0) return;
	
	float		u,v;
	OMediaRect  rect;

	u = pr.closer_hit.sub_info[0].u * canvas->get_width();
	v = pr.closer_hit.sub_info[0].v * canvas->get_height();
	
	
	OMediaCanvas	*srcCanv = [palette getCurrentBrush];
	if (!srcCanv) return;

	canvas->lock(omlf_Write|omlf_Read);
	
	if (!quiet && pr.closer_hit.element!=NULL)
	{
		if (mouseDownFirstDraw)
		{
			undoBufferFilled = true;
			memcpy(undoBuffer,canvas->get_pixels(),canvas->get_width()*canvas->get_height()*4);
		}
	
		int x,y;
	
		[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
		
		x = (int)(u-srcCanv->get_width()*0.5f);
		y = (int)(v-srcCanv->get_height()*0.5f);

		if (mouseDownFirstDraw || palette->currentTool==cosmic_Draw || (lastX==x && lastY==y))
			[self paintBrushAtX:x Y:y];
		else if (palette->currentTool==cosmic_ConnectedDraw)
		{
			if (lastPickedElement!=pr.closer_hit.element)
			{
				[self paintAcrossTilesToX:x toY:y element:pr.closer_hit.element];
			}
			else
			{
				//[self paintAcrossTilesToX:x toY:y element:pr.closer_hit.element];

				[self paintBrushAFromX:lastX Y:lastY toX:x toY:y];
			}
		}

		mouseDownFirstDraw = NO;
		lastX = x;
		lastY = y;
		lastPickedElement = pr.closer_hit.element;
		lastEX = (int)lastPickedElement->getx();
		lastEY = (int)lastPickedElement->gety();
	}

	canvas->unlock();
}


- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
    return YES;
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- (BOOL)becomeFirstResponder
{
    return YES;
}

- (BOOL)resignFirstResponder
{
    return YES;
}

-(IBAction)scaleSpeedChanged:(id)sender
{
	scaleSpeed = [palette->scaleSpeedSlider floatValue];
	[palette->scaleSpeedTextField setStringValue:[self floatToStringRound:scaleSpeed*100.0]];
	
	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)scrollSpeedChanged:(id)sender
{
	scrollSpeed = [palette->scrollSpeedSlider floatValue];
	[palette->scrollSpeedTextField setStringValue:[self floatToStringRound:scrollSpeed*100.0]];
	
	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)rotationSpeedChanged:(id)sender
{
	float f;
	
	f = [palette->rotationSpeedSlider floatValue];
	rotationSpeed = omd_Deg2AngleF(f);
	[palette->rotationSpeedTextField setStringValue:[self floatToStringRound:(f/1.5)*100]];
	
	f = [palette->fpsSlider floatValue];
	if (f<0.01) f=0.01;
	fpsFactor = f;
	[palette->fpsTextField setStringValue:[self floatToStringRound:f*100]];

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];

}

-(IBAction)updateScaleValues:(id)sender
{
	scaleSpeed = [palette->scaleSpeedSlider floatValue];
	scaleStart = [palette->scaleStartTextField floatValue]/100.0f;
	scaleEnd = [palette->scaleEndTextField floatValue]/100.0f;

	if (scaleStart<0) scaleStart = 0;
	else if (scaleStart>1.0f) scaleStart = 1.0f;

	if (scaleEnd<0) scaleEnd = 0;
	else if (scaleEnd>1.0f) scaleEnd = 1.0f;
	
	if (scaleEnd<scaleStart) scaleEnd = scaleStart;

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)updateScrollValues:(id)sender
{
	scrollSpeed = [palette->scrollSpeedSlider floatValue];

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)updateRotationValues:(id)sender
{
	float f;
	
	f = [palette->rotationSpeedSlider floatValue];
	rotationSpeed = omd_Deg2AngleF(f);	

	f = [palette->fpsSlider floatValue];
	if (f<0.01) f=0.01;
	fpsFactor = f;

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)updateManuallySetValues:(id)sender
{
	float   v;
	float   bsize = canvas->get_width();

	[self lock];

	if (sender==palette->manuallySetXTextField)
	{
		v = (([palette->manuallySetXTextField floatValue]/100.0f) * bsize);

		viewport->setx(v);
		[self realignViewPort];
	}
	else if (sender==palette->manuallySetYTextField)
	{
		v = (([palette->manuallySetYTextField floatValue]/100.0f) * bsize);

		viewport->sety(v);
		[self realignViewPort];
	}
	else if (sender==palette->manuallySetAngleTextField)
	{
		v = -[palette->manuallySetAngleTextField floatValue];
		viewport->set_anglez(omd_Deg2Angle(v));
	}
	else if (sender==palette->manuallySetZoomTextField)
	{
		v = COSMIC_MIN_Z + ([palette->manuallySetZoomTextField floatValue]/100.0f)*(COSMIC_MAX_Z-COSMIC_MIN_Z);
		if (v<COSMIC_MIN_Z) v = COSMIC_MIN_Z;
		else if (v>COSMIC_MAX_Z) v = COSMIC_MAX_Z;

		viewport->setz(v);
	}

	[self unlock];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}


-(IBAction)stopRotation:(id)sender
{
	rotationSpeed = 0;
	[palette->rotationSpeedSlider setFloatValue:0];	
	[palette->rotationSpeedTextField setFloatValue:0];

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
	
	[((CosmicDocument*)[[[self window] windowController] document]) updateChangeCount:NSChangeDone];
}

-(IBAction)stopScroll:(id)sender
{
	scrollSpeed = 0;
	[palette->scrollSpeedSlider setFloatValue:0];	
	[palette->scrollSpeedTextField setFloatValue:0];

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
	
	[((CosmicDocument*)[[[self window] windowController] document]) updateChangeCount:NSChangeDone];
}

-(IBAction)stopScale:(id)sender
{
	scaleSpeed = 0;
	[palette->scaleSpeedSlider setFloatValue:0];	
	[palette->scaleSpeedTextField setFloatValue:0];

	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
	[((CosmicDocument*)[[[self window] windowController] document]) updateChangeCount:NSChangeDone];
}

-(IBAction)resetAnimVectors:(id)sender
{
	viewport->place(0,0,0);
	viewport->set_angle(0,0,0);
	[self realignViewPort];
	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)clearBackground:(id)sender
{
    NSColor *color = [palette->clearColorWell color];
	OMediaFARGBColor	argb;

    [[color colorUsingColorSpaceName:@"NSDeviceRGBColorSpace"]  getRed:&argb.red green:&argb.green blue:&argb.blue alpha:&argb.alpha];

	[self lock];
	canvas->lock(omlf_Write|omlf_Read);	
	undoBufferFilled = true;
	memcpy(undoBuffer,canvas->get_pixels(),canvas->get_width()*canvas->get_height()*4);
	canvas->fill(argb);
	canvas->unlock();	
	[self unlock];

	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

-(IBAction)undo:(id)sender
{
	[self lock];
	unsigned long *tempBuffer = new unsigned long[canvas->get_width()*canvas->get_height()];
	
	canvas->lock(omlf_Write|omlf_Read);
	memcpy(tempBuffer,canvas->get_pixels(),canvas->get_width()*canvas->get_height()*4);
	memcpy(canvas->get_pixels(),undoBuffer,canvas->get_width()*canvas->get_height()*4);
	memcpy(undoBuffer,tempBuffer,canvas->get_width()*canvas->get_height()*4);
	canvas->unlock();
	
	delete [] tempBuffer;
	
	[self unlock];
}

- (BOOL)validateMenuItem:(NSMenuItem *)item 
{
    SEL action = [item action];
	
	if (recordMode) return NO;

    if (action == @selector(undo:)) return undoBufferFilled;
        
    if (action == @selector(exportTile:)) return YES;
    if (action == @selector(exportImage:)) return YES;
	if (action == @selector(exportMovie:)) return YES;
    if (action == @selector(mirrorDocument:)) 
	{
		CGDisplayCount num_displays;
	
		if (CGGetActiveDisplayList(0, NULL, &num_displays) != CGDisplayNoErr || num_displays <= 1) return NO;
		return YES;
	}

    return [super validateMenuItem:item];
}

-(IBAction)mirrorDocument:(id)sender
{
	[self lock];
	[CosmicMirror setViewToMirror:self];
	[self unlock];
}

- (OMediaCanvas*)glToCanvas:(BOOL)argbMode
{
	OMediaCanvas		*canvasBuffer;

	if (!renderTarget) return NULL;

	NSRect  rect = [self frame];
	int w = (int)rect.size.width, h = (int)rect.size.height;

	if (w<=0 || h<=0) return NULL;

	canvasBuffer = new OMediaCanvas;

	[self lock];
	
	[[self openGLContext] makeCurrentContext];

	canvasBuffer->create(w,h);
	canvasBuffer->lock(omlf_Read|omlf_Write);

	char	*buffer = new char[w*h*4],*dest,*src;

	glPixelStorei(GL_PACK_ROW_LENGTH,0);
	glPixelStorei(GL_PACK_SKIP_PIXELS,0);
	glPixelStorei(GL_PACK_ALIGNMENT,1);

	glReadPixels(0, 0,
				w,h,
				GL_RGBA,
				GL_UNSIGNED_BYTE, 
				buffer);

	[self unlock];

	dest = (char*)canvasBuffer->get_pixels();
	src = buffer + ((w<<2)*(h-1));
	int n = h,rowbytes = w<<2;

	while(n--)  // Flip lines
	{
		memcpy(dest,src,rowbytes);
		dest +=rowbytes;
		src -=rowbytes;
	}

	
	if (argbMode)
	{
		unsigned long *ldest = (unsigned long*)canvasBuffer->get_pixels();

		// Flip pixel guns
		
		int i,n = canvasBuffer->get_width()*canvasBuffer->get_height();

		for(i=0;i<n;i++)
		{
			unsigned long pix = *ldest;
			*ldest = (pix>>8UL)|0xFF000000UL;
			ldest++;
		}	
	}
	else
	canvasBuffer->fill_alpha(0xFF);

	delete [] buffer;

	canvasBuffer->unlock();

	return canvasBuffer;
}


- (void)exportImageDidEnd:(NSSavePanel *)sPanel returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
    if (returnCode == NSOKButton)
    {
		OMediaMoveableMem	*mb;
		NSData				*data;
		long				size;
		OMediaMemStream		stream;
	
		OMediaCanvas	*canvasBuffer = [self glToCanvas:NO];
		if (!canvasBuffer) return;

		canvasBuffer->png_export(stream);
		delete canvasBuffer;

		mb = stream.get_memory_block();
		size = mb->getsize();
    
		data = [NSData dataWithBytes:mb->lock() length:size];
		mb->unlock();

		[[NSFileManager defaultManager] createFileAtPath:[sPanel filename]
										contents:data 
										attributes:nil];

    }
}

-(IBAction)exportImage:(id)sender
{
    NSSavePanel *sPanel = [NSSavePanel savePanel];
    
    [sPanel setPrompt: NSLocalizedString(@"Export Image",@"")];
	[sPanel setRequiredFileType:@"png"];
    
    [sPanel beginSheetForDirectory:nil file:nil modalForWindow:[self window] 
            modalDelegate:self didEndSelector:@selector(exportImageDidEnd:returnCode:contextInfo:) contextInfo:nil];
}

- (void)exportTileDidEnd:(NSSavePanel *)sPanel returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
    if (returnCode == NSOKButton)
    {

		OMediaMoveableMem	*mb;
		NSData				*data;
		long				size;
		OMediaMemStream		stream;
		OMediaCanvas		canvasBuffer;

		if (!renderTarget) return;

		canvasBuffer.create(canvas->get_width(),canvas->get_height(),canvas);
		canvasBuffer.fill_alpha(0xFF);

		canvasBuffer.png_export(stream);

		mb = stream.get_memory_block();
		size = mb->getsize();
    
		data = [NSData dataWithBytes:mb->lock() length:size];
		mb->unlock();

		[[NSFileManager defaultManager] createFileAtPath:[sPanel filename]
										contents:data 
										attributes:nil];
    }
}


-(IBAction)exportTile:(id)sender
{
    NSSavePanel *sPanel = [NSSavePanel savePanel];
    
    [sPanel setPrompt: NSLocalizedString(@"Export Tile",@"")];
	[sPanel setRequiredFileType:@"png"];
    
    [sPanel beginSheetForDirectory:nil file:nil modalForWindow:[self window] 
            modalDelegate:self didEndSelector:@selector(exportTileDidEnd:returnCode:contextInfo:) contextInfo:nil];
}

-(IBAction)exportMovie:(id)sender
{
	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];

    [NSApp beginSheet:doc->exportMoviePanel
        modalForWindow:[self window]
        modalDelegate:nil
        didEndSelector:nil
		contextInfo:nil];
}

- (void)exportMovieDidEnd:(NSSavePanel *)sPanel returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
    if (returnCode == NSOKButton)
    {
		[self enterRecordMode:[sPanel filename]];
    }
}

-(IBAction)doExportMovie:(id)sender
{
	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];

	[NSApp endSheet:doc->exportMoviePanel];
	[doc->exportMoviePanel orderOut:self];

    NSSavePanel *sPanel = [NSSavePanel savePanel];
    
    [sPanel setPrompt: NSLocalizedString(@"Export Movie",@"")];
	[sPanel setRequiredFileType:@"mov"];
    
    [sPanel beginSheetForDirectory:nil file:nil modalForWindow:[self window] 
            modalDelegate:self didEndSelector:@selector(exportMovieDidEnd:returnCode:contextInfo:) contextInfo:nil];
}

-(IBAction)cancelExportMovie:(id)sender
{
	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];

	[NSApp endSheet:doc->exportMoviePanel];
	[doc->exportMoviePanel orderOut:self];
}

-(void)enterRecordMode:(NSString*)pathname
{
	NSSize	size;
	
	if (recordMode) return;

	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];
	recordMode = YES;
	
	[doc->infoTabView selectTabViewItemAtIndex:1];
	[doc->exportMovieProgress setDoubleValue:0.0];



	size.width = [[self window] frame].size.width;
	size.height = [[self window] frame].size.height;
	
	[[self window] setMaxSize:size];
	[[self window] setMinSize:size];

	float fps = exportMovieFps = [doc->exportFps selectedColumn]==0?20.0f:40.0f;
	float secs = [doc->exportMovieDuration floatValue];

	nframesToCompress = totalFramesToCompress = (int)(fps*secs);

	if (fps<=0 || secs<=0 || nframesToCompress<=0)
	{
		[self exitRecordMode];
		return;
	}

	wantedRate = 1.0/fps;
		
	if (![self initMovieCompression:pathname])
	{
		[self exitRecordMode];	
	}
}

-(void)recordNewFrame
{
	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];

	if (!recordMode) return;

	if (nframesToCompress<=0)
	{
		[self exitRecordMode];
		return;
	}

	[doc->exportMovieProgress setDoubleValue:(double)(totalFramesToCompress-nframesToCompress)/(double)totalFramesToCompress];

	if (renderLock)
	{
		if ([self lockFocusIfCanDraw])
		{		
			OMediaCanvas *frame;
				
			[self lock];
			[self periodical];
			[self render];
			[self unlockFocus];
			[self unlock];
			
			nframesToCompress--;
			
			frame = [self glToCanvas:YES];
			if (frame)
			{
				[self addFrameToMovieCompression:frame];
				delete frame;
			}
		}
	}

}

-(IBAction)abortExportMovie:(id)sender
{
	if (!recordMode) return;

	[self exitRecordMode];
}

-(void)exitRecordMode
{
	NSSize	size;

	[self finishMovieCompression];

	CosmicDocument *doc = (CosmicDocument*)[[[self window] windowController] document];

	wantedRate = 1.0/40.0;
	recordMode = NO;
	[doc->infoTabView selectTabViewItemAtIndex:0];

	size.width = FLT_MAX;
	size.height = FLT_MAX;

	[[self window] setMaxSize:size];

	size.width = 131;
	size.height = 86;
	[[self window] setMinSize:size];
}

- (void)resetCursorRects
{
	NSRect  rect = [self visibleRect];
	
	switch(palette->currentTool)
	{
		case cosmic_Draw:			[self addCursorRect:rect cursor:[NSCursor crosshairCursor]]; break;
		case cosmic_ConnectedDraw:  [self addCursorRect:rect cursor:[NSCursor crosshairCursor]]; break;
		case cosmic_Rotate:			[self addCursorRect:rect cursor:(mouseIsDown)?[NSCursor closedHandCursor]:[NSCursor openHandCursor]]; break;
		case cosmic_Scroll:			[self addCursorRect:rect cursor:(mouseIsDown)?[NSCursor closedHandCursor]:[NSCursor openHandCursor]]; break;
		case cosmic_Scale:			[self addCursorRect:rect cursor:(mouseIsDown)?[NSCursor closedHandCursor]:[NSCursor openHandCursor]]; break;
	}
}

- (void)setBounds:(NSRect)boundsRect
{
	[self setBounds:boundsRect];
	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
}

- (void)setBoundsSize:(NSSize)newSize
{
	[self setBoundsSize:newSize];
	[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];
}


- (void)updateUITimer:(NSTimer *)aTimer
{
	if ([[self window] isKeyWindow] && canvas)
	{
		float   bsize = canvas->get_width();
		float   x,y,z,a;

		[self lock];
		
		x = fmod(viewport->getx(),bsize)*100.0f/bsize;
		y = fmod(viewport->gety(),bsize)*100.0f/bsize;

		z = viewport->getz();
		if (palette->currentTool==cosmic_Scale)
		{
			z+=tempHandScale*200;
			if (z>COSMIC_MAX_Z) z=COSMIC_MAX_Z;
			else if (z<COSMIC_MIN_Z) z=COSMIC_MIN_Z;
		}
		
		z = (z-COSMIC_MIN_Z)*100.0f/(COSMIC_MAX_Z-COSMIC_MIN_Z);

		a = viewport->get_anglez();
		if (palette->currentTool==cosmic_Rotate && mouseIsDown) a-=handRotate;
		a = fmod(-omd_Angle2DegF(a),360.0);

		[self unlock];
		
		[palette->manuallySetXTextField setStringValue:[self floatToStringRound:x]];
		[palette->manuallySetYTextField setStringValue:[self floatToStringRound:y]];

		[palette->manuallySetAngleTextField setStringValue:[self floatToStringRound:a]];

		[palette->manuallySetZoomTextField setStringValue:[self floatToStringRound:z]];	
	}
}

- (void)mainThreadTimer:(NSTimer *)aTimer
{
	if (recordMode) [self recordNewFrame];

	if (keyWindowChanged && [[self window] isKeyWindow])
	{
	
		[palette->scaleSpeedTextField setStringValue:[self floatToStringRound:scaleSpeed*100.0]];
		[palette->scaleSpeedSlider setFloatValue:scaleSpeed];
		[palette->scaleStartTextField setStringValue:[self floatToStringRound:scaleStart*100.0f]];
		[palette->scaleEndTextField setStringValue:[self floatToStringRound:scaleEnd*100.0f]];
	
		[palette->scrollSpeedTextField setStringValue:[self floatToStringRound:scrollSpeed*100.0]];
		[palette->scrollSpeedSlider setFloatValue:scrollSpeed];

		[palette->rotationSpeedSlider setFloatValue:omd_Angle2DegF(rotationSpeed)];
		[palette->rotationSpeedTextField setStringValue:[self floatToStringRound:(omd_Angle2DegF(rotationSpeed)/1.5)*100]];

		[palette->fpsSlider setFloatValue:fpsFactor];
		[palette->fpsTextField setStringValue:[self floatToStringRound:fpsFactor*100]];

		[palette->scalePingpongCheckbox setIntValue:(int)scalePingpong];

		[((CosmicDocument*)[[[self window] windowController] document]) updateInfoText:self];

		keyWindowChanged = false;
	}
	else keyWindowChanged = true;
}

-(BOOL) newFilePath:(NSString *)filePath ToFSSpec:(FSSpec*)fsspec
{
	NSString			*parentDir;
    vector<unichar>		ustr;
    FSRef				fsref;
    Boolean				bres;
    OSErr				err = noErr;
	FSCatalogInfo		catalogInfo;

	if ([filePath length]==0) return false;

	parentDir = [filePath stringByDeletingLastPathComponent];

    ustr.resize([parentDir length]);
	[parentDir getCharacters:&ustr[0]];

    CFStringRef cfstr = CFStringCreateWithCharacters (NULL, &ustr[0],ustr.size());
    CFURLRef	cfurl = CFURLCreateWithFileSystemPath (NULL,cfstr,kCFURLPOSIXPathStyle,FALSE);
    
    bres = CFURLGetFSRef (cfurl,&fsref);
    if (bres)
    {
        err = FSGetCatalogInfo (&fsref,kFSCatInfoNodeID|kFSCatInfoVolume,&catalogInfo,NULL,fsspec,NULL);
    }

    CFRelease(cfurl);
    CFRelease(cfstr);
	
	if (err==noErr && bres)
	{
		string  filename = [[filePath lastPathComponent] lossyCString];
		unsigned char pstr[256];
		unsigned long len = filename.length();
		if (len>255) len = 255;
		pstr[0] = len;
		memcpy(pstr+1,filename.c_str(),len);

		FSMakeFSSpec (catalogInfo.volume, catalogInfo.nodeID, pstr,fsspec);
	}
    
    return (bres && err==noErr);
}


-(BOOL)initMovieCompression:(NSString*)path
{
	if (!recordMode) return NO;

	OSErr err = noErr;

	int	frameWidth = (int)[self frame].size.width,
		frameHeight = (int)[self frame].size.height;
	long	maxCompressedSize;
	Rect	rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = frameWidth;
	rect.bottom = frameHeight;

	FSSpec  mySpec;

	if (![self newFilePath:path ToFSSpec:&mySpec])
	{
		return NO;
	}
	
	movieRecord = new CosmicMovieCompress;
	movieRecord->frameRate = exportMovieFps;

	err = CreateMovieFile (&mySpec, 
                          FOUR_CHAR_CODE('TVOD'), // Creator is quicktime player
                          smCurrentScript, 
                          createMovieFileDeleteCurFile | createMovieFileDontCreateResFile,
                          &movieRecord->resRefNum, 
                          &movieRecord->theMovie );
	if (err==noErr)
	{
		// Create video track

		err = noErr;

		movieRecord->theTrack = NewMovieTrack (movieRecord->theMovie, 		/* movie specifier */
                                    FixRatio(frameWidth,1),  /* width */
                                    FixRatio(frameHeight,1), /* height */
								  kNoVolume);  /* trackVolume */

		if (GetMoviesError()==noErr)
		{
			// Create media for track
		
			movieRecord->theMedia = NewTrackMedia (movieRecord->theTrack,			/* track identifier */
									VideoMediaType,		/* type of media */
									kVideoTimeScale, 	/* time coordinate system */
									nil,				/* data reference - use the file that is associated with the movie  */
									0);					/* data reference type */

			if (GetMoviesError()==noErr)
			{
				err = BeginMediaEdits (movieRecord->theMedia);
				if (err==noErr)
				{
				
				   err = NewGWorld (&movieRecord->theGWorld,	/* pointer to created gworld */	
							32,		/* pixel depth */
							&rect, 		/* bounds */
							nil, 			/* color table */
							nil,			/* handle to GDevice */ 
							(GWorldFlags)0);	/* flags */

				
				   err = NewGWorld (&movieRecord->previousGWorld,	/* pointer to created gworld */	
							32,		/* pixel depth */
							&rect, 		/* bounds */
							nil, 			/* color table */
							nil,			/* handle to GDevice */ 
							(GWorldFlags)0);	/* flags */

					// Lock the pixels
					LockPixels (GetGWorldPixMap(movieRecord->theGWorld));
					LockPixels (GetGWorldPixMap(movieRecord->previousGWorld));

					err = GetMaxCompressionSize(GetGWorldPixMap(movieRecord->theGWorld),		/* Handle to the source image */
										&rect,									/* bounds */
										0,										/* let ICM choose depth */
										codecNormalQuality,							/* desired image quality */ 
										kAnimationCodecType,								/* compressor type */ 
										(CompressorComponent)anyCodec,			/* compressor identifier */
										&maxCompressedSize);					/* returned size */

					// Create a new handle of the right size for our compressed image data
					movieRecord->compressedData = NewHandle(maxCompressedSize);

					MoveHHi( movieRecord->compressedData );
					HLock( movieRecord->compressedData );
					movieRecord->compressedDataPtr = *movieRecord->compressedData;

					// Create a handle for the Image Description Structure
					movieRecord->imageDesc = (ImageDescriptionHandle)NewHandle(4);

					// Now begin sequence
					err = CompressSequenceBegin (&movieRecord->imageSequence,
											GetGWorldPixMap(movieRecord->theGWorld),
											GetGWorldPixMap(movieRecord->previousGWorld),
											&rect,&rect,
											32,
											kAnimationCodecType,
											anyCodec,
											codecNormalQuality,
											codecNormalQuality,
											5,		/* keyFrameRate*/
											NULL,   /* ctable */
											codecFlagUpdatePreviousComp,
											movieRecord->imageDesc);
					
					if (err==noErr) return YES;						
				}
			}
		}

		CloseMovieFile (movieRecord->resRefNum);
	}
	
	delete movieRecord;
	movieRecord = NULL;

	return NO;
}
	
-(void)finishMovieCompression
{
	if (!movieRecord) return;

	CDSequenceEnd(movieRecord->imageSequence);
	
	// Deallocate our previously alocated handles and GWorld
	if (movieRecord->imageDesc)
		DisposeHandle ((Handle)movieRecord->imageDesc);
	
	if (movieRecord->compressedData)
		DisposeHandle (movieRecord->compressedData);
	
	if (movieRecord->theGWorld)
	{
		UnlockPixels (GetGWorldPixMap(movieRecord->theGWorld));
		DisposeGWorld (movieRecord->theGWorld);
	}
	
	if (movieRecord->previousGWorld)
	{
		UnlockPixels (GetGWorldPixMap(movieRecord->previousGWorld));
		DisposeGWorld (movieRecord->previousGWorld);		
	}

	EndMediaEdits (movieRecord->theMedia);
	InsertMediaIntoTrack (movieRecord->theTrack,			/* track specifier */
				0,							/* track start time */
				0,							/* media start time */
				GetMediaDuration(movieRecord->theMedia), /* media duration */
				fixed1);					/* media rate ((Fixed) 0x00010000L) */

	short resId = movieInDataForkResID;
	OSErr	err;

	err = AddMovieResource (movieRecord->theMovie, movieRecord->resRefNum, &resId, nil);

	CloseMovieFile (movieRecord->resRefNum);		
	movieRecord = NULL;	
}

-(void)addFrameToMovieCompression:(OMediaCanvas*)mcanvas
{
	CGrafPtr oldPort;
	GDHandle oldGDeviceH;
	GWorldPtr srcGWorld;
	Rect			srcRect;
	long			dataSize;
	UInt8			similarity;
	OSErr			err;

	GetGWorld(&oldPort, &oldGDeviceH);
    SetGWorld(movieRecord->theGWorld, nil);

	if ((int)mcanvas->get_width()<=0 || (int)mcanvas->get_height()<=0) return;
			
	srcRect.left = 0;
	srcRect.top = 0;
	srcRect.right = (int)mcanvas->get_width();
	srcRect.bottom = (int)mcanvas->get_height();

	mcanvas->lock(omlf_Write|omlf_Read);

	err = QTNewGWorldFromPtr (&srcGWorld, k32ARGBPixelFormat, &srcRect, 
                                            NULL, NULL, 0, 
                                            mcanvas->get_pixels(),
                                            mcanvas->get_width()*4);	
			
	CopyBits (( BitMapPtr ) *GetGWorldPixMap(srcGWorld),( BitMapPtr ) *GetGWorldPixMap(movieRecord->theGWorld),
						&srcRect,&srcRect,
						srcCopy,NULL);

	DisposeGWorld(srcGWorld);

	err = CompressSequenceFrame (movieRecord->imageSequence,
									GetGWorldPixMap(movieRecord->theGWorld),
									&srcRect,
									codecFlagUpdatePreviousComp,
									movieRecord->compressedDataPtr,
									&dataSize,
									&similarity,
									NULL);

	// Add sample data and a description to a media
	err = AddMediaSample(movieRecord->theMedia,					/* media specifier */ 
			movieRecord->compressedData,							/* handle to sample data - dataIn */
			0,										/* specifies offset into data reffered to by dataIn handle */
			dataSize,					/* number of bytes of sample data to be added */ 
			(int)(kVideoTimeScale/movieRecord->frameRate),				/* frame duration  */
			(SampleDescriptionHandle)movieRecord->imageDesc,		/* sample description handle */ 
			1,										/* number of samples */
			(similarity!=0)?mediaSampleNotSync:0,						/* control flag indicating self-contained samples */
			nil);									/* returns a time value where sample was insterted */
			
	mcanvas->unlock();
				
	SetGWorld (oldPort, oldGDeviceH);	
}

-(IBAction)pingPongChanged:(id)sender
{
	scalePingpong = [palette->scalePingpongCheckbox intValue]!=0;
	[[[[self window] windowController] document] updateChangeCount:NSChangeDone];
}

@end

